import logging
import json
from typing import Optional, Dict, Generator
import requests
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Interface:
    def __init__(self, raw_json: Dict):
        self.id = raw_json.get("id", None)
        self.name = raw_json.get("name", None)
        self.alias = raw_json.get("alias", None)
        self.mac = raw_json.get("mac", None)
        self.ip = raw_json.get("ip", None)
        self.mask = raw_json.get("mask", None)
        self.link = raw_json.get("link", None)
        self.speed = raw_json.get("speed", None)
        self.duplex = raw_json.get("duplex", None)
        self.tx_packets = raw_json.get("tx_packets", None)
        self.rx_packets = raw_json.get("rx_packets", None)
        self.tx_bytes = raw_json.get("tx_bytes", None)
        self.rx_bytes = raw_json.get("rx_bytes", None)
        self.tx_errors = raw_json.get("tx_errors", None)
        self.rx_errors = raw_json.get("rx_errors", None)

    def __str__(self):
        return f"Interface({self.name}, {self.ip})"

    def __repr__(self):
        return self.__str__()


class Resource:
    def __init__(self, raw_json):
        self.cpu = raw_json.get("cpu")[0]["current"]
        self.sessions = raw_json.get("session")[0]["current"]
        self.memory = raw_json.get("mem")[0]["current"]
        self.disk = raw_json.get("disk")[0]["current"]
        self.disk_log_rate = raw_json.get("disk_lograte")[0]["current"]

    def __str__(self):
        return f"Resource(CPU: {self.cpu}, Memory: {self.memory}, Sessions: {self.sessions}, Disk: {self.disk})"

    def __repr__(self):
        return self.__str__()


class Proxy:
    def __init__(self, raw_json):
        self.p2name: str = raw_json.get("p2name")
        self.status: str = raw_json.get("status")
        self.p2serial: int = raw_json.get("p2serial")
        self.expire: int = raw_json.get("expire", 0)
        self.incoming_bytes: int = raw_json.get("incoming_bytes", 0)
        self.outgoing_bytes: int = raw_json.get("outgoing_bytes", 0)

    def __str__(self):
        return f"Proxy({self.p2name}: {self.status})"

    def __repr__(self):
        return self.__str__()


class Tunnel:
    def __init__(self, raw_json: Dict):
        self.name: str = raw_json.get("name")
        self.comments: str = raw_json.get("comments")
        self.type: str = raw_json.get("type")
        self.incoming_bytes: int = raw_json.get("incoming_bytes")
        self.outgoing_bytes: int = raw_json.get("outgoing_bytes")
        self.rgwy: str = raw_json.get("rgwy")

        self.proxies = []
        for proxy in raw_json["proxyid"]:
            self.proxies.append(Proxy(proxy))

    def __str__(self):
        return f"Tunnel({self.name})"

    def __repr__(self):
        return self.__str__()


class InterfaceAvailable:
    def __init__(self, raw_json: Dict):
        self.name = raw_json.get("name", None)
        self.status = raw_json.get("status", None)

    def __str__(self):
        return f"InterfaceAvailable(name: {self.name}, Status: {self.status})"

    def __repr__(self):
        return self.__str__()


class Fortigate:
    def __init__(self, url: str, user: str, password: str, log: logging.Logger, proxies=None):
        self.base_url: str = url
        self.user: str = user
        self.password: str = password
        self.log = log

        self.api_url: str = f"{self.base_url}/api/v2"
        self.session: requests.Session = requests.Session()
        self.proxies = proxies

    def login(self):
        url = f"{self.base_url}/logincheck"
        params = {"username": self.user, "secretkey": self.password}
        self.make_request(url, params=params, method="POST")

    def logout(self):
        url = f"{self.base_url}/logout"
        self.make_request(url, method="POST")

    def make_request(self, url, params=None, headers=None, method="GET") -> Optional[requests.Response]:
        self.log.info(f"Making request to '{url}'")
        with self.session as s:
            try:
                r = s.request(
                    method,
                    url,
                    params=params,
                    verify=False,
                    headers=headers,
                    timeout=10,
                    proxies=self.proxies,
                )
                if r.status_code >= 400:
                    self.log.error(f"Error calling {url}: {r}")
                    return None
                return r
            except Exception as e:
                self.log.error(f"Error calling {url}: {e}")
                return None

    def _get_system_resource_usage(self) -> Dict:
        url = f"{self.api_url}/monitor/system/resource/usage"
        params = {"scope": "global"}
        return self.make_request(url, params=params).json()

    def _get_system_interface_select(self) -> Dict:
        url = f"{self.api_url}/monitor/system/interface/select"
        params = {"scope": "global"}
        return self.make_request(url, params=params).json()

    def _get_system_available_interface_select(self) -> Dict:
        url = f"{self.api_url}/monitor/system/available-interfaces/select"
        return self.make_request(url).json()

    def get_system_status_select(self) -> Dict:
        url = f"{self.api_url}/monitor/system/status/select"
        params = {"scope": "global"}
        return self.make_request(url, params=params).json()

    def _get_system_config_revision_select(self) -> Dict:
        url = f"{self.api_url}/monitor/system/config-revision/select"
        params = {"scope": "global"}
        return self.make_request(url, params=params).json()

    def _get_system_global_resources_select(self) -> Dict:
        url = f"{self.api_url}/monitor/system/global-resources/select"
        params = {"scope": "global"}
        return self.make_request(url, params=params).json()

    def _get_vpn_ipsec_select(self) -> Dict:
        url = f"{self.api_url}/monitor/vpn/ipsec/select"
        params = {"scope": "global"}
        return self.make_request(url, params=params).json()

    def _get_vpn_ssl_select(self) -> Dict:
        url = f"{self.api_url}/monitor/vpn/ssl/select"
        params = {"scope": "global"}
        return self.make_request(url, params=params).json()

    @property
    def interfaces(self) -> Generator[Interface, None, None]:
        response = self._get_system_interface_select()
        for name, details in response["results"].items():
            yield Interface(details)

    @property
    def resources(self) -> Resource:
        response = self._get_system_resource_usage()
        return Resource(response["results"])

    @property
    def tunnels(self) -> Generator[Tunnel, None, None]:
        response = self._get_vpn_ipsec_select()
        for details in response["results"]:
            yield Tunnel(details)

    @property
    def interface_available(self) -> Generator[InterfaceAvailable, None, None]:
        response = self._get_system_available_interface_select()
        for details in response["results"]:
            yield InterfaceAvailable(details)


def dump_to_file(filename: str, contents: dict):
    with open(filename, "w") as f:
        json.dump(contents, f, indent=2)


def main():
    log = logging.getLogger(__name__)
    log.setLevel(logging.INFO)
    st = logging.StreamHandler()
    log.addHandler(st)
    f = Fortigate("https://ec2-3-94-144-93.compute-1.amazonaws.com", "user", "user", log)
    f.login()

    dump_to_file("mock/resource.json", f._get_system_resource_usage())
    # dump_to_file("mock/interface.json", f.get_system_interface_select())
    # dump_to_file("mock/status.json", f.get_system_status_select())
    # dump_to_file("mock/vpnipsec.json", f.get_vpn_ipsec_select())
    # dump_to_file("mock/vpnssl.json", f.get_vpn_ssl_select())
    # dump_to_file("mock/config_revision.json", f.get_system_config_revision_select())

    for t in f.tunnels:
        print(t)
        for p in t.proxies:
            print(p)
        print()


if __name__ == "__main__":
    main()
