import logging
import re
import socket
from urllib.parse import urlparse, quote

from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import ConfigException

from python_ping import ping
from fortigate_api import Fortigate

log = logging.getLogger(__name__)


class FortigatePlugin(RemoteBasePlugin):

    def initialize(self, **kwargs):

        parsed_url = urlparse(self.config.get("url"))
        valid_ip = r"^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$"
        valid_hostname = r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"

        target = str(parsed_url.netloc.split(":")[0])
        if not re.match(valid_ip, target) and not re.match(valid_hostname, target):
            raise ConfigException(f"Invalid test_target: {target}, must be a valid IP or hostname")

    def query(self, **kwargs):

        proxies = None
        if self.config.get("proxy_host"):
            username = self.config.get("proxy_username")
            password = self.config.get("proxy_password")
            host = self.config.get("proxy_host")
            port = self.config.get("proxy_port")

            proxies = {
                "https": f"http://{username}:{quote(password)}@{host}:{port}",
                "http": f"http://{username}:{quote(password)}@{host}:{port}",
            }

        f = Fortigate(
            self.config.get("url"),
            self.config.get("username"),
            self.config.get("password"),
            log,
            proxies=proxies,
        )

        group = self.topology_builder.create_group(self.config.get("group"), self.config.get("group"))
        device = group.create_device(self.config.get("alias"))

        parsed_url = urlparse(self.config.get("url"))
        ping_result = ping(parsed_url.netloc)
        log.info(f"Ping result: {ping_result.as_dict()}")
        device.absolute(
            "ping_availability",
            (100 - ping_result.packet_loss_rate) if ping_result.packet_loss_rate is not None else 0,
        )

        if ping_result.rtt_avg is not None:
            device.absolute("ping_rtt", ping_result.rtt_avg)

        try:
            f.login()

            status = f.get_system_status_select()
            device.report_property("Version", status.get("version", "unknown"))
            device.report_property("Serial", status.get("serial", "unknown"))
            device.absolute("connectivity", 100)

            ip = None
            try:
                ip = socket.gethostbyname(parsed_url.netloc)
            except Exception:
                try:
                    ip = socket.inet_aton(parsed_url.netloc)
                except Exception:
                    self.logger.warning(f"Could not determine the IP for {parsed_url.netloc}")

            if ip is not None:
                device.add_endpoint(ip)

            resources = f.resources
            device.absolute("cpu", resources.cpu)
            device.absolute("memory", resources.memory)
            device.absolute("sessions", resources.sessions)
            device.absolute("disk", resources.disk)
            device.absolute("disk_log_rate", resources.disk_log_rate)

            for interface in f.interfaces:
                device.per_second("bytes_in", interface.rx_bytes, dimensions={"Interface": interface.name})
                device.per_second(
                    "bytes_out",
                    interface.tx_bytes,
                    dimensions={"Interface": interface.name},
                )
                device.relative(
                    "errors_in",
                    interface.rx_errors,
                    dimensions={"Interface": interface.name},
                )
                device.relative(
                    "errors_out",
                    interface.tx_errors,
                    dimensions={"Interface": interface.name},
                )

            for tunnel in f.tunnels:
                device.per_second(
                    "tun_bytes_in",
                    tunnel.incoming_bytes,
                    dimensions={"Tunnel": tunnel.name},
                )
                device.per_second(
                    "tun_bytes_out",
                    tunnel.outgoing_bytes,
                    dimensions={"Tunnel": tunnel.name},
                )
                for proxy in tunnel.proxies:
                    device.per_second(
                        "tun_proxy__bytes_in",
                        proxy.incoming_bytes,
                        dimensions={"Tunnel": tunnel.name, "Proxy": proxy.p2name},
                    )
                    device.per_second(
                        "tun_proxy_bytes_out",
                        proxy.outgoing_bytes,
                        dimensions={"Tunnel": tunnel.name, "Proxy": proxy.p2name},
                    )

                    device.state_metric(
                        "tun_proxy_status",
                        proxy.status.upper(),
                        dimensions={"Tunnel": tunnel.name, "Proxy": proxy.p2name},
                    )

            regexes = [line.strip() for line in self.config.get("interfaces_filter", "").splitlines() if line.strip()]
            if not regexes:
                regexes = [".*"]

            for interface in f.interface_available:
                if interface.status:
                    # check if the interface name matches any of the regexes
                    match = [re.match(regex, interface.name) for regex in regexes]
                    if not any(match):
                        self.logger.info(f"Skipping interface {interface.name} because it didn't match any regex: {regexes}")
                        continue
                    self.logger.info(f"Adding interface {interface.name} because it matched regex: {regexes}, reporting status: {interface.status.upper()}")
                    device.state_metric(
                        "interface_status",
                        interface.status.upper(),
                        dimensions={"Interface": interface.name}
                    )

        except Exception as e:
            log.exception(f"Could not connect or obtain metrics: {e}")
            device.absolute("connectivity", 0)
        finally:
            try:
                f.logout()
            except Exception as e:
                log.exception(f"Logout failed: {e}")
